/*
	skyeye_options.c - skyeye config file options' functions
	Copyright (C) 2003 Skyeye Develop Group
	for help please send mail to <skyeye-developer@lists.sf.linuxforum.net> 
	
	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 
 
*/
/* 08/20/2003   add log option function
				chenyu
   4/02/2003	add net option function
 * 				walimis <walimi@peoplemail.com.cn>
 * 3/22/2003 	add cpu, mem_num, mem_bank, arch, dummy option function
 *				walimis <walimi@peoplemail.com.cn> 		
 * 10/24/2005	add dbct test speed function
 *				teawater <c7code-uc@yahoo.com.cn>
 * */
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include "skyeye_options.h"
#include "skyeye_arch.h"
#include "skyeye_config.h"

#include "skyeye_arch.h"
#include "skyeye_config.h"

extern FILE *skyeye_logfd;


int
split_param (const char *param, char *name, char *value)
{
	const char *src = param;
	char *dst = name;

	while (*src && (*src != '='))
		*dst++ = *src++;
	*dst = '\0';

	if (*src == '\0') {
		value = '\0';
		return -1;
	}

	strcpy (value, src + 1);
	return 0;
}


/* we need init some options before read the option file.
 * now put them here.
 * */
//chy 2004-12-05 for eclipse
//#define SKYEYE4ECLIPSE
#ifdef SKYEYE4ECLIPSE
extern char *inferior_io_terminal;
int nouart;
#endif
//------------------------------
int
skyeye_option_init (skyeye_config_t * config)
{
//chy 2004-12-05 for eclipse
#ifdef SKYEYE4ECLIPSE
	int fd;
#endif
//------------------------------                        

	config->uart.fd_in = 0;	/*stdin */
	config->uart.fd_out = 1;	/*stdout */
	/*ywc 2005-04-01 */
	config->no_dbct = 1;	/*default, dbct is off */
	//teawater add for new tb manage function 2005.07.10----------------------------
	config->tb_tbt_size = 0;
	config->tb_tbp_size = TB_TBP_DEFAULT;

//chy 2004-12-05 for eclipse
#ifdef SKYEYE4ECLIPSE
	nouart = 0;
	if (inferior_io_terminal != NULL) {
		nouart = 1;
		printf ("SKYEYE: inferior_io_terminal is %s \n",
			inferior_io_terminal);
		if ((fd = open (inferior_io_terminal, O_RDWR)) == -1)
			SKYEYE_DBG ("Error when open device %s",
				    inferior_io_terminal);
		skyeye_config.uart.fd_in = fd;
		skyeye_config.uart.fd_out = fd;
	}
#endif
//------------------------------                        

}
int
do_dummy_option (skyeye_option_t * this_option, int num_params,
		 const char *params[])
{
	return 0;
};

/* parse "int" parameters. e.g. int=16:17*/
int
get_interrupts (char value[], u32 * interrupts)
{
	char *cur = value;
	char *next = value;
	int i = 0, end = 0;

	while ((*next != ':') && (*next != 0))
		next++;

	while (*cur != 0) {
		if (*next != 0) {
			*next = '\0';
		}
		else
			end = 1;
		interrupts[i] = strtoul (cur, NULL, 0);
		//printf("%s:%s\n", __FUNCTION__, cur);
		i++;
		if ((i > 4) || end == 1)
			return 0;
		cur = ++next;
		while ((*next != ':') && (*next != 0))
			next++;
	}
	return 0;
}

/* setup device option.
 * all device options may have common parameters. Here we handle these common parameters.
 * */
int
setup_device_option (char *option_name, void *dev_option,
		     int num_params, const char *params[])
{
	char name[MAX_PARAM_NAME], value[MAX_PARAM_NAME];
	struct common_config conf;
	int i;

	memset (&conf, 0, sizeof (conf));
	conf.type = NULL;
	conf.name = NULL;

	for (i = 0; i < num_params; i++) {
		if (split_param (params[i], name, value) < 0)
			SKYEYE_ERR
				("Error: \"%s\" has wrong parameter \"%s\".\n",
				 option_name, name);
		if (!strncmp ("type", name, strlen (name))) {
			conf.type = strdup (value);
		}
		else if (!strncmp ("name", name, strlen (name))) {
			conf.name = strdup (value);
		}
		else if (!strncmp ("base", name, strlen (name))) {
			conf.base = strtoul (value, NULL, 0);
		}
		else if (!strncmp ("size", name, strlen (name))) {
			conf.size = strtoul (value, NULL, 0);
		}
		else if (!strncmp ("int", name, strlen (name))) {
			get_interrupts (value, conf.interrupts);
		}
	}
	setup_device (option_name, &conf, dev_option, skyeye_config.mach);
	if (conf.type)
		free (conf.type);
	if (conf.name)
		free (conf.name);
	return 0;

}

/*defined in skyeye_arch.c*/
extern arch_config_t *skyeye_archs[];

int
do_arch_option (skyeye_option_t * this_option, int num_params,
		const char *params[])
{
	int i;
	arch_config_t *arch = skyeye_config.arch;
	char *default_arch = "arm";

	for (i = 0; i < MAX_SUPP_ARCH; i++) {
		if (skyeye_archs[i] == NULL)
			continue;
		if (!strncmp
		    (params[0], skyeye_archs[i]->arch_name, MAX_PARAM_NAME)) {
			skyeye_config.arch = skyeye_archs[i];
			SKYEYE_INFO ("arch: %s\n",
				     skyeye_archs[i]->arch_name);
			return 0;
		}
	}
	SKYEYE_ERR
		("Error: Unknowm architecture name \"%s\" or you use low version of skyeye?\n",
		 params[0]);
	return -1;
}

int
do_cpu_option (skyeye_option_t * this_option, int num_params,
	       const char *params[])
{
	int ret;

	if (skyeye_config.arch == NULL) {
		/* If we don't set arch, we use "arm" as default. */
		char *default_arch = "arm";
		int i;
		for (i = 0; i < MAX_SUPP_ARCH; i++) {
			if (skyeye_archs[i] == NULL)
				continue;
			if (!strncmp
			    (default_arch, skyeye_archs[i]->arch_name,
			     MAX_PARAM_NAME)) {
				skyeye_config.arch = skyeye_archs[i];
				SKYEYE_INFO ("arch: %s\n",
					     skyeye_archs[i]->arch_name);
			}
		}
		if (skyeye_config.arch == NULL) {
			SKYEYE_ERR
				("ERROR: No arch option found! Maybe you use low version of skyeye?\n");
			skyeye_exit (-1);
		}
	}
	ret = skyeye_config.arch->parse_cpu (params);
	if (ret < 0)
		SKYEYE_ERR ("Error: Unkonw cpu name \"%s\"\n", params[0]);
	return ret;
}

int
do_mach_option (skyeye_option_t * this_option, int num_params,
		const char *params[])
{
	int ret;
	machine_config_t *mach = skyeye_config.mach;
	ret = skyeye_config.arch->parse_mach (mach, params);
	if (ret < 0) {
		SKYEYE_ERR ("Error: Unkonw mach name \"%s\"\n", params[0]);
	}
	return ret;
}

int
do_mem_bank_option (skyeye_option_t * this_option, int num_params,
		    const char *params[])
{
	int ret;
	ret = skyeye_config.arch->parse_mem (num_params, params);
	if (ret < 0) {
		SKYEYE_ERR ("Error: Unkonw mem bank name \"%s\"\n",
			    params[0]);
	}
	return ret;
}


int
do_net_option (skyeye_option_t * this_option, int num_params,
	       const char *params[])
{
	char name[MAX_PARAM_NAME], value[MAX_PARAM_NAME];
	int i;
	struct net_option net_opt;
	unsigned char mac_addr[6];
	unsigned char hip[4];
	unsigned char *maddr, *ip;

	for (i = 0; i < num_params; i++) {
		if (split_param (params[i], name, value) < 0) {
			SKYEYE_ERR ("Error: %s has wrong parameter \"%s\".\n",
				    this_option->name, name);
			return -1;
		}
		if (!strncmp ("mac", name, strlen (name))) {
			sscanf (value, "%x:%x:%x:%x:%x:%x", &mac_addr[0],
				&mac_addr[1], &mac_addr[2], &mac_addr[3],
				&mac_addr[4], &mac_addr[5]);
			memcpy (net_opt.macaddr, mac_addr, 6);

		}
		else if (!strncmp ("hostip", name, strlen (name))) {
			/* FIXME: security problem, don't use sscanf later */
			sscanf (value, "%d.%d.%d.%d", &hip[0], &hip[1],
				&hip[2], &hip[3]);
			memcpy (net_opt.hostip, hip, 4);

		}
		else if (!strncmp ("ethmod", name, strlen (name))) {
			if (!strncmp ("linux", value, strlen (value))) {
				net_opt.ethmod = NET_MOD_LINUX;

			}
			else if (!strncmp ("tuntap", value, strlen (value))) {
				net_opt.ethmod = NET_MOD_TUNTAP;

			}
			else if (!strncmp ("win", value, strlen (value))) {
				net_opt.ethmod = NET_MOD_WIN;

			}
			else if (!strncmp ("vnet", value, strlen (value))) {
				net_opt.ethmod = NET_MOD_VNET;
			}
		}
	}
	maddr = net_opt.macaddr;
	ip = net_opt.hostip;
	printf ("ethmod num=%d, mac addr=%x:%x:%x:%x:%x:%x, hostip=%d.%d.%d.%d\n", net_opt.ethmod, maddr[0], maddr[1], maddr[2], maddr[3], maddr[4], maddr[5], ip[0], ip[1], ip[2], ip[3]);
	setup_device_option (this_option->name, (void *) &net_opt, num_params,
			     params);

	return 0;
}

/*added by ksh in 2004-11-26 ,energy option*/
int
do_energy_option (skyeye_option_t * this_option, int num_params,
		  const char *params[])
{
	char name[MAX_PARAM_NAME], value[MAX_PARAM_NAME];
	int i, fd;

	for (i = 0; i < num_params; i++) {
		if (split_param (params[i], name, value) < 0)
			SKYEYE_ERR
				("Error: energy has wrong parameter \"%s\".\n",
				 name);
		if (!strncmp ("state", name, strlen (name))) {
			if (!strncmp ("on", value, strlen (value))) {
				skyeye_config.energy.energy_prof = 1;
				SKYEYE_INFO
					("energy info: turn on energy!\n");
			}
			else {
				skyeye_config.energy.energy_prof = 0;
				SKYEYE_INFO
					("energy info: turn off energy!\n");
			}
		}
		else if (!strncmp ("filename", name, strlen (name))) {
			strcpy (skyeye_config.energy.filename, value);
			skyeye_config.energy.filename[strlen (value)] = '\0';
		}
		else if (!strncmp ("logfile", name, strlen (name))) {
			strcpy (skyeye_config.energy.logfile, value);
			skyeye_config.energy.logfile[strlen (value)] = '\0';
		}
	}
	return 0;

}



/*uart option*/
int
do_uart_option (skyeye_option_t * this_option, int num_params,
		const char *params[])
{
	char name[MAX_PARAM_NAME], value[MAX_PARAM_NAME];
	int i, fd;

//chy 2004-12-05 for eclipse
#ifdef SKYEYE4ECLIPSE
	if (nouart)
		return 0;
#endif
//----------------------------
	for (i = 0; i < num_params; i++) {
		if (split_param (params[i], name, value) < 0)
			SKYEYE_ERR
				("Error: uart has wrong parameter \"%s\".\n",
				 name);
		if (!strncmp ("fd_in", name, strlen (name))) {
			if ((fd = open (value, O_RDWR)) == -1)
				SKYEYE_DBG ("Error when open device %s",
					    value);
			skyeye_config.uart.fd_in = fd;
			SKYEYE_INFO
				("uart info: input device is %s, fd is %d.\n",
				 value, fd);
		}
		else if (!strncmp ("fd_out", name, strlen (name))) {
			if ((fd = open (value, O_RDWR)) == -1)
				SKYEYE_DBG ("Error when open device %s",
					    value);
			skyeye_config.uart.fd_out = fd;
			SKYEYE_INFO
				("uart info: output device is %s, fd is %d.\n",
				 value, fd);
		}
	}
	return 0;

}


/* Anthony Lee 2006-09-07 */
static int lcd_mod_is_valid(int mod)
{
#ifdef GTK_LCD
	if(mod == LCD_MOD_GTK) return 1;
#endif
#ifdef WIN32_LCD
	if(mod == LCD_MOD_WIN32) return 1;
#endif
	return 0;
}


/* Anthony Lee 2006-09-07 */
static int lcd_valid_mod(void)
{
#ifdef GTK_LCD
	return LCD_MOD_GTK;
#else
#ifdef WIN32_LCD
	return LCD_MOD_WIN32;
#else
	return LCD_MOD_NONE;
#endif
#endif
}


/*chy 2004-03-11 lcd option*/
int
do_lcd_option (skyeye_option_t * this_option, int num_params,
	       const char *params[])
{
	char name[MAX_PARAM_NAME], value[MAX_PARAM_NAME];
	struct lcd_option lcd_opt;
	int i, fd;

	for (i = 0; i < num_params; i++) {
		if (split_param (params[i], name, value) < 0)
			SKYEYE_ERR
				("Error: lcd has wrong parameter \"%s\".\n",
				 name);
		if (!strncmp ("mod", name, strlen (name))) {
			if (!strncmp ("gtk", value, strlen (value))) {
				lcd_opt.mod = LCD_MOD_GTK;
			}
			else if (!strncmp ("qt", value, strlen (value))) {
				lcd_opt.mod = LCD_MOD_QT;
			}
			else if (!strncmp ("x", value, strlen (value))) {
				lcd_opt.mod = LCD_MOD_X;
			}
			else if (!strncmp ("sdl", value, strlen (value))) {
				lcd_opt.mod = LCD_MOD_SDL;
			}
			else if (!strncmp ("win32", value, strlen (value))) {
				lcd_opt.mod = LCD_MOD_WIN32;
			}

			/* Anthony Lee 2006-09-07 */
			if(lcd_mod_is_valid(lcd_opt.mod) == 0)
			{
				SKYEYE_INFO("*** Note: The module of LCD not compiled in.\n***       SkyEye will try the valid value.\n");
				lcd_opt.mod = lcd_valid_mod();
			}
		}
	}

	setup_device_option (this_option->name, (void *) &lcd_opt, num_params,
			     params);
}

/*ywc 2005-04-01 dbct option*/
int
do_dbct_option (skyeye_option_t * this_option, int num_params,
		const char *params[])
{
	char name[MAX_PARAM_NAME], value[MAX_PARAM_NAME];
	int i, fd;

	for (i = 0; i < num_params; i++) {
		if (split_param (params[i], name, value) < 0)
			SKYEYE_ERR
				("Error: dbct has wrong parameter \"%s\".\n",
				 name);
		if (!strncmp ("state", name, strlen (name))) {
			if (!strncmp ("on", value, strlen (value))) {
				skyeye_config.no_dbct = 0;
#ifndef NO_DBCT
				SKYEYE_INFO ("dbct info: turn on dbct!\n");
#else
				SKYEYE_INFO
					("dbct info: Note: DBCT not compiled in. This option will be ignored\n");
#endif
			}
			else {
				skyeye_config.no_dbct = 1;
				SKYEYE_INFO ("dbct info: turn off dbct!\n");
			}
		}
		//teawater add for new tb manage function 2005.07.10----------------------------
		if (!strncmp ("tbt", name, strlen (name))) {
			skyeye_config.tb_tbt_size = strtoul (value, NULL, 0);
		}
		if (!strncmp ("tbp", name, strlen (name))) {
			skyeye_config.tb_tbp_size = strtoul (value, NULL, 0);
		}
	}
	return 0;
}

//chy:2003-08-20 do_log option
int
do_log_option (skyeye_option_t * this_option, int num_params,
	       const char *params[])
{
	char name[MAX_PARAM_NAME], value[MAX_PARAM_NAME];
	int i, fd, logon, memlogon;
	unsigned long long start, end, length;
	/*2004-08-09 chy init skyeye_config.log */
	skyeye_config.log.log_fd = 0;
	skyeye_config.log.logon = 0;
	skyeye_config.log.memlogon = 0;
	skyeye_config.log.start = 0;
	skyeye_config.log.end = 0;
	skyeye_config.log.length = 0;

	for (i = 0; i < num_params; i++) {
		if (split_param (params[i], name, value) < 0)
			SKYEYE_ERR
				("log_info: Error: log has wrong parameter \"%s\".\n",
				 name);
		if (!strncmp ("logon", name, strlen (name))) {
			sscanf (value, "%d", &logon);
			if (logon != 0 && logon != 1)
				SKYEYE_ERR
					("log_info: Error logon value %d\n",
					 logon);
			if (logon == 1) {
				SKYEYE_INFO ("log_info: log is on.\n");
			}
			else {
				SKYEYE_INFO ("log_info: log is off.\n");
			}
			skyeye_config.log.logon = logon;
		}
		else if (!strncmp ("memlogon", name, strlen (name))) {
			sscanf (value, "%d", &memlogon);
			if (memlogon != 0 && memlogon != 1)
				SKYEYE_ERR
					("log_info: Error logon value %d\n",
					 memlogon);
			if (memlogon == 1) {
				SKYEYE_INFO ("log_info: memory klog is on.\n");
			}
			else {
				SKYEYE_INFO ("log_info: memory log is off.\n");
			}
			skyeye_config.log.memlogon = memlogon;
		}
		else if (!strncmp ("logfile", name, strlen (name))) {
			if ((skyeye_logfd = fopen (value, "w+")) == NULL) {
				//SKYEYE_DBG("SkyEye Error when open log file %s\n", value);
				perror ("SkyEye: Error when open log file:  ");
				skyeye_exit (-1);
			}
			skyeye_config.log.log_fd = skyeye_logfd;
			SKYEYE_INFO ("log_info:log file is %s, fd is 0x%x\n",
				     value, skyeye_logfd);
		}
		else if (!strncmp ("start", name, strlen (name))) {
			start = strtoul (value, NULL, 0);
			skyeye_config.log.start = start;
			SKYEYE_INFO ("log_info: log start clock %llu\n",
				     start);
		}
		else if (!strncmp ("end", name, strlen (name))) {
			end = strtoul (value, NULL, 0);
			skyeye_config.log.end = end;
			SKYEYE_INFO ("log_info: log end clock %llu\n", end);
		}
		else if (!strncmp ("length", name, strlen (name))) {
			sscanf (value, "%llu", &length);
			skyeye_config.log.length = length;
			SKYEYE_INFO ("log_info: log instr length %llu\n",
				     length);
		}
	}
	return 0;
}


/*
 * we can add this option using:
 * step_disassemble:on[ON|1] for open this option
 * step_disassemble:off[OFF|0] for disable this option 
 * oyangjian add here
 */
int
do_step_disassemble_option (skyeye_option_t * this_option, int num_params,
			    const char *params[])
{
	int i;

	for (i = 0; i < num_params; i++) {
		printf ("step_disassemble state:%s\n", params[0]);
		if (!params[0]) {
			SKYEYE_ERR ("Error :usage: step_disassemble:on[off]");
			return -1;
		}
		if (!strncmp (params[0], "on", 2)
		    || !strncmp (params[0], "ON", 2)) {
			skyeye_config.can_step_disassemble = 1;
			return 0;
		}
		else if (!strncmp (params[0], "off", 3)
			 || !strncmp (params[0], "OFF", 3)) {

			skyeye_config.can_step_disassemble = 0;
			return 0;
		}

	}
	SKYEYE_ERR ("Error: Unkonw cpu name \"%s\"\n", params[0]);
	return -1;
}

int
do_flash_option (skyeye_option_t * this_option, int num_params,
		 const char *params[])
{
	char name[MAX_PARAM_NAME], value[MAX_PARAM_NAME];
	struct flash_option flash_opt;
	int i, fd;

	for (i = 0; i < num_params; i++) {
		if (split_param (params[i], name, value) < 0)
			SKYEYE_ERR
				("Error: flash has wrong parameter \"%s\".\n",
				 name);
		/*
		   if (!strncmp ("mod", name, strlen (name)))
		   {
		   if (!strncmp ("gtk", value, strlen (value)))
		   {
		   lcd_opt.mod = LCD_MOD_GTK;
		   }
		   else if (!strncmp ("qt", value, strlen (value)))
		   {
		   lcd_opt.mod = LCD_MOD_QT;
		   }
		   else if (!strncmp ("x", value, strlen (value)))
		   {
		   lcd_opt.mod = LCD_MOD_X;
		   }
		   else if (!strncmp ("sdl", value, strlen (value)))
		   {
		   lcd_opt.mod = LCD_MOD_SDL;
		   }
		   }
		 */
	}
	setup_device_option (this_option->name, (void *) &flash_opt,
			     num_params, params);
}

//teawater add DBCT_TEST_SPEED 2005.10.04---------------------------------------
#ifdef DBCT_TEST_SPEED
int
do_dbct_test_speed_sec(struct skyeye_option_t *this_opion, int num_params, const char *params[])
{
	if (num_params != 1) {
		goto error_out;
	}
	errno = 0;
	skyeye_config.dbct_test_speed_sec = strtol(params[0], NULL, 10);
	if (errno == ERANGE) {
		goto error_out;
	}
	printf("dbct_test_speed_sec %ld\n", skyeye_config.dbct_test_speed_sec);

error_out:
	SKYEYE_ERR("Error :usage: step_disassemble: dbct_test_speed_sec\n");
	return(-1);
}
#endif	//DBCT_TEST_SPEED
//AJ2D--------------------------------------------------------------------------
